"""
Main module for spotdl. Exports version and main function.
"""

from spotdl.console import console_entry_point
from spotdl._version import __version__

if __name__ == "__main__":
    console_entry_point()



"""
AZLyrics lyrics module.
"""

from typing import List, Optional

from bs4 import BeautifulSoup

import requests

from spotdl.providers.lyrics.base import LyricsProvider


class AzLyrics(LyricsProvider):
    """
    AZLyrics lyrics provider class.
    """

    def get_lyrics(self, name: str, artists: List[str], **_) -> Optional[str]:
        """
        Try to get lyrics from azlyrics

        ### Arguments
        - name: The name of the song.
        - artists: The artists of the song.

        ### Returns
        - The lyrics of the song or None if no lyrics were found.
        """

        # Join every artist by comma in artists
        artist_str = ", ".join(artist for artist in artists if artist)

        song_name = name.replace(" ", "+").lower()
        song_artists = artist_str.replace(" ", "+").lower()
        song_artists = song_artists.replace(",", "%2C")

        url = f"https://search.azlyrics.com/search.php?q={song_name}+{artists}"

        response = requests.get(url, headers=self.headers)
        soup = BeautifulSoup(response.content, "html.parser")

        td_tags = soup.find_all("td")
        if len(td_tags) == 0:
            return None

        result = td_tags[0]

        a_tags = result.find_all("a", href=True)
        if len(a_tags) != 0:
            lyrics_url = a_tags[0]["href"]
        else:
            return None

        if lyrics_url.strip() == "":
            return None

        response = requests.get(lyrics_url, headers=self.headers)
        soup = BeautifulSoup(response.content, "html.parser")

        # Find all divs that don't have a class
        div_tags = soup.find_all("div", class_=False, id_=False)

        # Find the div with the longest text
        lyrics_div = sorted(div_tags, key=lambda x: len(x.text))[-1]

        lyrics = lyrics_div.get_text()

        # Remove the 3 first new lines
        lyrics = lyrics[3:]

        return lyrics



"""
Sync module for the console.
"""

import json

from pathlib import Path
from typing import List, Optional

from spotdl.download.downloader import Downloader
from spotdl.utils.search import parse_query
from spotdl.utils.formatter import create_file_name
from spotdl.utils.m3u import create_m3u_file
from spotdl.types.song import Song


def sync(
    query: List[str],
    downloader: Downloader,
    save_path: Optional[Path] = None,
    m3u_file: Optional[str] = None,
    **_,
) -> None:
    """
    Sync function for the console.
    It will download the songs and remove the ones that are no longer
    present in the playlists/albums/etc


    ### Arguments
    - query: list of strings to search for.
    - downloader: Already initialized downloader instance.
    - save_path: Path to save the songs to.
    - m3u_file: Path to the file to save the metadata to.
    """

    downloader.save_file = None

    # Query and save file
    # Create initial sync file
    if query and save_path:
        if any(req for req in query if req.endswith(".spotdl")):
            # If the query contains a .spotdl file, and we are about to create
            # .spotdl file, raise an error.
            raise ValueError(
                "Cannot create a sync file with a .spotdl file in the query."
            )

        # Parse the query
        songs_list = parse_query(query, downloader.threads)

        # Create sync file
        with open(save_path, "w", encoding="utf-8") as save_file:
            json.dump(
                {
                    "type": "sync",
                    "query": query,
                    "songs": [song.json for song in songs_list],
                },
                save_file,
                indent=4,
                ensure_ascii=False,
            )

        # Perform initial download
        downloader.download_multiple_songs(songs_list)

        if m3u_file:
            create_m3u_file(
                m3u_file, songs_list, downloader.output, downloader.output_format, False
            )

        return None

    # If the query is a single file, download it
    if len(query) == 1 and query[0].endswith(".spotdl") and not save_path:
        # Load the sync file
        with open(query[0], "r", encoding="utf-8") as sync_file:
            sync_data = json.load(sync_file)

        # Verify the sync file
        if sync_data.get("type") != "sync":
            raise ValueError("Sync file is not a valid sync file.")

        # Parse the query
        songs_list = parse_query(sync_data["query"], downloader.threads)

        # Get all the old files based on the songs from sync file
        old_files = [
            create_file_name(Song(**song), downloader.output, downloader.output_format)
            for song in sync_data["songs"]
        ]

        # Get all output file names
        new_files = [
            create_file_name(song, downloader.output, downloader.output_format)
            for song in songs_list
        ]

        # Get all files that are no longer in the song lists
        to_delete = set(old_files) - set(new_files)

        # Delete all files that are no longer in the song lists
        for file in to_delete:
            if file.exists():
                downloader.progress_handler.log(f"Deleting {file}")
                file.unlink()
            else:
                downloader.progress_handler.debug(f"{file} does not exist.")

        to_download = []
        for song in songs_list:
            song_path = create_file_name(
                song, downloader.output, downloader.output_format
            )
            if Path(song_path).exists():
                if downloader.overwrite == "force":
                    downloader.progress_handler.log(f"Overwriting {song.display_name}")
                    to_download.append(song)
            else:
                to_download.append(song)

        if len(to_download) == 0:
            downloader.progress_handler.log("Nothing to do...")
            return None

        if m3u_file:
            create_m3u_file(
                m3u_file,
                songs_list,
                downloader.output,
                downloader.output_format,
                False,
            )

        with open(query[0], "w", encoding="utf-8") as save_file:
            json.dump(
                {
                    "type": "sync",
                    "query": sync_data["query"],
                    "songs": [song.json for song in songs_list],
                },
                save_file,
                indent=4,
                ensure_ascii=False,
            )

        downloader.download_multiple_songs(to_download)

        return None

    raise ValueError(
        "Wrong combination of arguments. "
        "Either provide a query and a save path. Or a single sync file in the query"
    )



"""
Genius Lyrics module.
"""

from typing import List, Optional

import requests

from bs4 import BeautifulSoup
from spotdl.providers.lyrics.base import LyricsProvider


class Genius(LyricsProvider):
    """
    Genius lyrics provider class.
    """

    def get_lyrics(self, name: str, artists: List[str], **_) -> Optional[str]:
        """
        Try to get lyrics from genius

        ### Arguments
        - name: The name of the song.
        - artists: The artists of the song.

        ### Returns
        - The lyrics of the song or None if no lyrics were found.
        """

        try:
            headers = {
                "Authorization": "Bearer "
                "alXXDbPZtK1m2RrZ8I4k2Hn8Ahsd0Gh_o076HYvcdlBvmc0ULL1H8Z8xRlew5qaG",
            }

            headers.update(self.headers)

            artist_str = ", ".join(
                artist for artist in artists if artist.lower() not in name.lower()
            )

            search_response = requests.get(
                "https://api.genius.com/search",
                params={"q": f"{name} {artist_str}"},
                headers=headers,
            )

            song_id = search_response.json()["response"]["hits"][0]["result"]["id"]

            song_response = requests.get(
                f"https://api.genius.com/songs/{song_id}", headers=headers
            )

            song_url = song_response.json()["response"]["song"]["url"]

            counter = 0
            soup = None
            while counter > 5:
                genius_page_response = requests.get(song_url, headers=self.headers)

                print(genius_page_response.text)

                if not genius_page_response.ok:
                    counter += 1
                    continue

                soup = BeautifulSoup(
                    genius_page_response.text.replace("<br/>", "\n"), "html.parser"
                )
                break

            if soup is None:
                return None

            lyrics_div = soup.select_one("div.lyrics")

            if lyrics_div is not None:
                return lyrics_div.get_text().strip()

            lyrics_containers = soup.select("div[class^=Lyrics__Container]")
            lyrics = "\n".join(con.get_text() for con in lyrics_containers)
            return lyrics.strip()
        except Exception:
            return None



"""
Artist module for retrieving artist data from Spotify.
"""

from dataclasses import dataclass
from typing import Any, Dict, List, Set

from slugify import slugify

from spotdl.types.song import SongList
from spotdl.types.song import Song
from spotdl.types.album import Album
from spotdl.utils.spotify import SpotifyClient


class ArtistError(Exception):
    """
    Base class for all exceptions related to artists.
    """


@dataclass(frozen=True)
class Artist(SongList):
    """
    Artist class.
    Contains all the information about an artist.
    Frozen to prevent accidental modification.
    """

    genres: List[str]
    albums: List[Album]

    @classmethod
    def from_url(cls, url: str) -> "Artist":
        """
        Creates an Artist object from a URL.

        ### Arguments
        - url: The URL of the artist.

        ### Returns
        - The Artist object.
        """

        if "open.spotify.com" not in url or "artist" not in url:
            raise ArtistError(f"Invalid URL: {url}")

        metadata = Artist.get_metadata(url)
        album_urls = cls.get_albums(url)

        tracks: List[Song] = []
        albums: List[Album] = []

        # get artist tracks
        # same as above, but for tracks
        known_tracks: Set[str] = set()
        if len(album_urls) < 1:
            raise ArtistError(
                "Couldn't get albums, check if you have passed correct artist id"
            )

        # get all tracks from all albums
        # ignore duplicates
        urls = []
        for album_url in album_urls:
            album = Album.from_url(album_url)
            albums.append(album)
            for track in album.songs:
                track_name = slugify(track.name)  # type: ignore
                if track_name not in known_tracks:
                    tracks.append(track)
                    urls.append(track.url)
                    known_tracks.add(track_name)

        return cls(
            **metadata,
            songs=tracks,
            albums=albums,
            urls=urls,
        )

    @staticmethod
    def get_urls(url: str) -> List[str]:
        """
        Get urls for all songs for artist.

        ### Arguments
        - url: The URL of the artist.

        ### Returns
        - List of urls for all songs for artist.
        """

        albums = Artist.get_albums(url)

        urls = []
        for album in albums:
            urls.extend(Album.get_urls(album))

        return urls

    @classmethod
    def create_basic_list(cls, url: str) -> "Artist":
        """
        Create a basic list with only the required metadata and urls.

        ### Arguments
        - url: The url of the list.

        ### Returns
        - The SongList object.
        """

        metadata = Artist.get_metadata(url)
        urls = Artist.get_urls(url)

        return cls(**metadata, urls=urls, songs=[], albums=[])

    @staticmethod
    def get_albums(url: str) -> List[str]:
        """
        Returns a list with album urls.

        ### Arguments
        - url: The URL of the artist.

        ### Returns
        - List of album urls.
        """

        # query spotify for artist details
        spotify_client = SpotifyClient()

        artist_albums = spotify_client.artist_albums(url, album_type="album,single")

        albums: List[str] = []

        # get artist albums and remove duplicates
        # duplicates can occur if the artist has the same album available in
        # different countries
        known_albums: Set[str] = set()
        if artist_albums is not None:
            for album in artist_albums["items"]:
                albums.append(album["external_urls"]["spotify"])
                known_albums.add(slugify(album["name"]))  # type: ignore

            # Fetch all artist albums
            while artist_albums and artist_albums["next"]:
                artist_albums = spotify_client.next(artist_albums)
                if artist_albums is None:
                    break

                for album in artist_albums["items"]:
                    album_name = slugify(album["name"])  # type: ignore

                    if album_name in known_albums:
                        albums.extend([item["uri"] for item in artist_albums["items"]])

                        known_albums.add(album_name)

        return albums

    @staticmethod
    def get_metadata(url: str) -> Dict[str, Any]:
        """
        Get metadata for artist.

        ### Arguments
        - url: The URL of the artist.

        ### Returns
        - Dict with metadata for artist.
        """

        # query spotify for artist details
        spotify_client = SpotifyClient()

        # get artist info
        raw_artist_meta = spotify_client.artist(url)

        if raw_artist_meta is None:
            raise ArtistError(
                "Couldn't get metadata, check if you have passed correct artist id"
            )

        return {
            "name": raw_artist_meta["name"],
            "genres": raw_artist_meta["genres"],
            "url": url,
        }



"""
Module for creating m3u content and writing it to a file.
"""

from typing import List
from spotdl.types.song import Song
from spotdl.utils.formatter import create_file_name


def create_m3u_content(
    song_list: List[Song], template: str, file_extension: str, short: bool = False
) -> str:
    """
    Create m3u content and return it as a string.

    ### Arguments
    - song_list: the list of songs
    - template: the template to use
    - file_extension: the file extension to use
    - short: whether to use the short version of the template

    ### Returns
    - the m3u content as a string
    """

    text = ""
    for song in song_list:
        text += str(create_file_name(song, template, file_extension, short)) + "\n"

    return text


def create_m3u_file(
    file_name: str,
    song_list: List[Song],
    template: str,
    file_extension: str,
    short: bool = False,
) -> str:
    """
    Create the m3u file.

    ### Arguments
    - file_name: the file name to use
    - song_list: the list of songs
    - template: the template to use
    - file_extension: the file extension to use
    - short: whether to use the short version of the template

    ### Returns
    - the m3u content as a string
    """

    m3u_content = create_m3u_content(song_list, template, file_extension, short)

    with open(file_name, "w", encoding="utf-8") as m3u_file:
        m3u_file.write(m3u_content)

    return m3u_content



"""
Module that holds the entry point for the console.
"""

import sys
import json
import signal
import logging
import cProfile
import pstats

from rich.traceback import install
from rich.console import Console

from spotdl.console.download import download
from spotdl.console.sync import sync
from spotdl.console.save import save
from spotdl.download import Downloader
from spotdl.providers.audio.base import AudioProviderError
from spotdl.providers.audio.ytmusic import YouTubeMusic
from spotdl.utils.config import DEFAULT_CONFIG, ConfigError, get_config
from spotdl.utils.ffmpeg import (
    FFmpegError,
    download_ffmpeg,
    get_local_ffmpeg,
    is_ffmpeg_installed,
)
from spotdl.utils.config import get_config_file
from spotdl.utils.github import check_for_updates
from spotdl.utils.arguments import parse_arguments
from spotdl.utils.spotify import SpotifyClient, SpotifyError
from spotdl.download.downloader import DownloaderError


OPERATIONS = {
    "download": download,
    "sync": sync,
    "save": save,
}


def entry_point():
    """
    Console entry point for spotdl. This is where the magic happens.
    """

    # Don't log too much
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("spotipy").setLevel(logging.NOTSET)
    logging.getLogger("asyncio").setLevel(logging.WARNING)

    # Create a console
    console = Console()

    # Install rich traceback handler
    install(show_locals=False, extra_lines=1)

    # Create config file if it doesn't exist
    if get_config_file().is_file() is False:
        config_path = get_config_file()
        with open(config_path, "w", encoding="utf-8") as config_file:
            json.dump(DEFAULT_CONFIG, config_file, indent=4)

    if getattr(sys, "frozen", False) and len(sys.argv) == 1:
        # If the application is frozen, we check for ffmpeg
        # if it's not present download it create config file
        if is_ffmpeg_installed() is False:
            download_ffmpeg()

        try:
            get_config()
        except ConfigError:
            config_path = get_config_file()
            with open(config_path, "w", encoding="utf-8") as config_file:
                json.dump(DEFAULT_CONFIG, config_file, indent=4)

    # Download ffmpeg if the `--download-ffmpeg` flag is passed
    # This is done before the argument parser so it doesn't require `operation`
    # and `query` to be passed. Exit after downloading ffmpeg
    if "--download-ffmpeg" in sys.argv:
        if get_local_ffmpeg() is not None or is_ffmpeg_installed():
            overwrite_ffmpeg = input(
                "FFmpeg is already installed. Do you want to overwrite it? (y/N): "
            )

            if overwrite_ffmpeg.lower() == "y":
                local_ffmpeg = download_ffmpeg()

                if local_ffmpeg.is_file():
                    print(
                        f"FFmpeg successfully downloaded to {local_ffmpeg.absolute()}"
                    )
                else:
                    print("FFmpeg download failed")
        else:
            print("Downloading FFmpeg...")
            download_path = download_ffmpeg()

            if download_path.is_file():
                print(f"FFmpeg successfully downloaded to {download_path.absolute()}")
            else:
                print("FFmpeg download failed")

        return None

    # Generate the config file if it doesn't exist
    # or overwrite the current config file if the `--overwrite-config` flag is passed
    # This is done before the argument parser so it doesn't requires `operation`
    # and `query` to be passed. exit after downloading ffmpeg
    if "--generate-config" in sys.argv:
        config_path = get_config_file()
        if config_path.exists():
            overwrite_config = input("Config file already exists. Overwrite? (y/N): ")

            if overwrite_config.lower() != "y":
                print("Exiting...")
                return None

        with open(config_path, "w", encoding="utf-8") as config_file:
            json.dump(DEFAULT_CONFIG, config_file, indent=4)

        print(f"Config file generated at {config_path}")

        return None

    # Get information about the current version and display it
    # Exit after displaying the version
    if "--check-for-updates" in sys.argv:
        check_for_updates()

        return None

    # Parse the arguments
    arguments = parse_arguments()

    # Get the config file
    # It will automatically load if the `load_config` is set to True
    # in the config file
    config = {}
    if arguments.config or (
        get_config_file().exists() and get_config().get("load_config")
    ):
        config = get_config()

    # Create settings dict
    # If the setting in config file is the same as
    # the default setting, it will be ignored
    # and argument value will be used
    settings = {}
    for key, value in DEFAULT_CONFIG.items():
        argument_key_val = arguments.__dict__.get(key)
        config_key_val = config.get(key)

        if argument_key_val is not None and config_key_val == value:
            settings[key] = argument_key_val
        elif argument_key_val is None and config_key_val != value:
            settings[key] = config_key_val
        else:
            settings[key] = value

    # Check if ffmpeg is installed
    if is_ffmpeg_installed(settings["ffmpeg"]) is False:
        raise FFmpegError(
            "FFmpeg is not installed. Please run `spotdl --download-ffmpeg` to install it, "
            "or `spotdl --ffmpeg /path/to/ffmpeg` to specify the path to ffmpeg."
        )

    if "youtube-music" in settings["audio_providers"]:
        # Check if we are getting results from YouTube Music
        ytm = YouTubeMusic(settings)
        test_results = ytm.get_results("a")
        if len(test_results) == 0:
            raise AudioProviderError(
                "Could not connect to YouTube Music API. Use a VPN or other audio provider."
            )

    # Initialize spotify client
    SpotifyClient.init(
        client_id=settings["client_id"],
        client_secret=settings["client_secret"],
        user_auth=settings["user_auth"],
        cache_path=settings["cache_path"],
        no_cache=settings["no_cache"],
        open_browser=not settings["headless"],
    )

    # If the application is frozen start web ui
    # or if the operation is `web`
    if (
        getattr(sys, "frozen", False)
        and len(sys.argv) == 1
        or arguments.operation == "web"
    ):
        from spotdl.console.web import (  # pylint: disable=C0415,C0410,W0707,W0611
            web,
        )

        # Don't log too much when running web ui & default logging argument
        if arguments.__dict__.get("log_level") is None:
            settings["log_level"] = "CRITICAL"

        # Start web ui
        web(settings)

        return None

    # Check if save file is present and if it's valid
    if isinstance(settings["save_file"], str) and not settings["save_file"].endswith(
        ".spotdl"
    ):
        raise DownloaderError("Save file has to end with .spotdl")

    if arguments.query and "saved" in arguments.query and not settings["user_auth"]:
        raise SpotifyError(
            "You must be logged in to use the saved query. \
Log in by adding the --user-auth flag"
        )

    # Initialize the downloader
    # for download, load and preload operations
    downloader = Downloader(
        audio_providers=settings["audio_providers"],
        lyrics_providers=settings["lyrics_providers"],
        ffmpeg=settings["ffmpeg"],
        bitrate=settings["bitrate"],
        ffmpeg_args=settings["ffmpeg_args"],
        output_format=settings["format"],
        save_file=settings["save_file"],
        threads=settings["threads"],
        output=settings["output"],
        overwrite=settings["overwrite"],
        search_query=settings["search_query"],
        cookie_file=settings["cookie_file"],
        log_level=settings["log_level"],
        simple_tui=settings["simple_tui"],
        restrict=settings["restrict"],
        print_errors=settings["print_errors"],
        sponsor_block=settings["sponsor_block"],
    )

    def graceful_exit(_signal, _frame):
        downloader.progress_handler.close()
        sys.exit(0)

    signal.signal(signal.SIGINT, graceful_exit)
    signal.signal(signal.SIGTERM, graceful_exit)

    try:
        # Pick the operation to perform
        # based on the name and run it!
        OPERATIONS[arguments.operation](
            query=arguments.query,
            save_path=settings["save_file"],
            preload=settings["preload"],
            downloader=downloader,
            m3u_file=settings["m3u"],
        )
    except Exception:
        downloader.progress_handler.close()

        console.print_exception(show_locals=False, extra_lines=1)

        sys.exit(1)

    downloader.progress_handler.close()

    return None


def console_entry_point():
    """
    Wrapper around `entry_point` so we can profile the code
    """

    if "--profile" in sys.argv:
        with cProfile.Profile() as profile:
            entry_point()

        stats = pstats.Stats(profile)
        stats.sort_stats(pstats.SortKey.TIME)

        # Use snakeviz to visualize the profile
        stats.dump_stats("spotdl.profile")
    else:
        entry_point()



"""
Youtube module for downloading and searching songs.
"""

from typing import Any, Dict, List, Optional

from pytube import YouTube as PyTube, Search
from rapidfuzz import fuzz
from slugify import slugify

from spotdl.utils.formatter import create_song_title, create_search_query
from spotdl.providers.audio.base import AudioProvider
from spotdl.types import Song


class YouTube(AudioProvider):
    """
    YouTube audio provider class
    """

    def search(self, song: Song) -> Optional[str]:
        """
        Search for a video on YouTube.

        ### Arguments
        - song: The song to search for.

        ### Returns
        - The url of the best match or None if no match was found.
        """

        if self.search_query:
            search_query = create_search_query(
                song, self.search_query, False, None, True
            )
        else:
            # if isrc is not None then we try to find song with it
            if song.isrc:
                isrc_results = self.get_results(song.isrc)

                if isrc_results and len(isrc_results) == 1:
                    isrc_result = isrc_results[0]

                    if isrc_result and isrc_result.watch_url is not None:
                        return isrc_result.watch_url

            search_query = create_song_title(song.name, song.artists).lower()

        # Query YTM by songs only first, this way if we get correct result on the first try
        # we don't have to make another request to ytmusic api that could result in us
        # getting rate limited sooner
        results = self.get_results(search_query)

        if results is None:
            return None

        if self.filter_results:
            ordered_results = {results[0].watch_url: 100}
        else:
            # Order results
            ordered_results = self.order_results(results, song)

        # No matches found
        if len(ordered_results) == 0:
            return None

        result_items = list(ordered_results.items())

        # Sort results by highest score
        sorted_results = sorted(result_items, key=lambda x: x[1], reverse=True)

        # Return the first result
        return sorted_results[0][0]

    @staticmethod
    def get_results(
        search_term: str, *_args, **_kwargs
    ) -> Optional[List[PyTube]]:  # pylint: disable=W0221
        """
        Get results from YouTube

        ### Arguments
        - search_term: The search term to search for.
        - args: Unused.
        - kwargs: Unused.

        ### Returns
        - A list of YouTube results if found, None otherwise.
        """

        return Search(search_term).results

    def order_results(self, results: List[PyTube], song: Song) -> Dict[str, Any]:
        """
        Filter results based on the song's metadata.

        ### Arguments
        - results: The results to order.
        - song: The song to order for.

        ### Returns
        - The ordered results.
        """

        # Assign an overall avg match value to each result
        links_with_match_value = {}

        # Slugify song title
        slug_song_name = slugify(song.name)
        slug_song_title = slugify(
            create_song_title(song.name, song.artists)
            if not self.search_query
            else create_search_query(song, self.search_query, False, None, True)
        )

        for result in results:
            # Skip results without id
            if result.video_id is None:
                continue

            # Slugify some variables
            slug_result_name = slugify(result.title)
            sentence_words = slug_song_name.replace("-", " ").split(" ")

            # Check for common words in result name
            common_word = any(
                word != "" and word in slug_result_name for word in sentence_words
            )

            # skip results that have no common words in their name
            if not common_word:
                continue

            # Find artist match
            artist_match_number = 0.0

            # Calculate artist match for each artist
            # in the song's artist list
            for artist in song.artists:
                artist_match_number += fuzz.partial_token_sort_ratio(
                    slugify(artist), slug_result_name
                )

            # skip results with artist match lower than 70%
            artist_match = artist_match_number / len(song.artists)
            if artist_match < 70:
                continue

            # Calculate name match
            name_match = fuzz.partial_token_sort_ratio(
                slug_result_name, slug_song_title
            )

            # Drop results with name match lower than 50%
            if name_match < 50:
                continue

            # Calculate time match
            time_match = (
                100 - (result.length - song.duration**2) / song.duration * 100
            )

            average_match = (artist_match + name_match + time_match) / 3

            # the results along with the avg Match
            links_with_match_value[result.watch_url] = average_match

        return links_with_match_value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spotdl',
 'spotdl.console',
 'spotdl.download',
 'spotdl.providers',
 'spotdl.providers.audio',
 'spotdl.providers.lyrics',
 'spotdl.types',
 'spotdl.utils']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.10.0,<5.0.0',
 'fastapi>=0.78.0,<0.79.0',
 'mutagen>=1.45.1,<2.0.0',
 'nest_asyncio>=1.5.4,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-slugify>=6.1.1,<7.0.0',
 'pytube>=12.1.0,<13.0.0',
 'rapidfuzz>=2.0.11,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich>=12.4.4,<13.0.0',
 'spotipy>=2.19.0,<3.0.0',
 'uvicorn>=0.17.1,<0.18.0',
 'yt-dlp>=2022.5.18,<2023.0.0',
 'ytmusicapi>=0.22.0,<0.23.0']

entry_points = \
{'console_scripts': ['spotdl = spotdl:console_entry_point']}

setup_kwargs = {
    'name': 'spotdl',
    'version': '4.0.0rc1',
    'description': 'Download your Spotify playlists and songs along with album art and metadata',
    'long_description': '<!--- mdformat-toc start --slug=github --->\n\n<!--- \n!!! IF EDITING THE README, ENSURE TO COPY THE WHOLE FILE TO index.md in `/docs/`\n--->\n\n<div align="center">\n\n# spotDL v4\n\nDownload your Spotify playlists and songs along with album art and metadata\n\n[![MIT License](https://img.shields.io/apm/l/atomic-design-ui.svg?style=flat-square&color=44CC11)](https://github.com/spotDL/spotify-downloader/blob/master/LICENSE)\n[![PyPI version](https://img.shields.io/pypi/pyversions/spotDL?color=%2344CC11&style=flat-square)](https://pypi.org/project/spotdl/)\n![GitHub commits since latest release (by date)](https://img.shields.io/github/commits-since/spotDL/spotify-downloader/latest?color=44CC11&style=flat-square)\n[![PyPi downloads](https://img.shields.io/pypi/dw/spotDL?label=downloads@pypi&color=344CC11&style=flat-square)](https://pypi.org/project/spotdl/)\n![Contributors](https://img.shields.io/github/contributors/spotDL/spotify-downloader?style=flat-square)\n[![Discord](https://img.shields.io/discord/771628785447337985?label=discord&logo=discord&style=flat-square)](https://discord.gg/xCa23pwJWY)\n\n</div>\n\n> A new and improved version of spotDL: still the fastest, easiest and most accurate command-line music downloader\n\n---\n\n**[Read the documentation on ReadTheDocs!](http://spotdl.rtfd.io/)**\n\n---\n\n## Prerequisites\n\n- [Visual C++ 2019 redistributable](https://docs.microsoft.com/pl-pl/cpp/windows/latest-supported-vc-redist?view=msvc-170#visual-studio-2015-2017-2019-and-2022) **(on Windows)**\n- Python 3.7 or above (added to PATH)\n\n> **_YouTube Music must be available in your country for spotDL to work. This is because we use YouTube Music to filter search results. You can check if YouTube Music is available in your country, by visiting [YouTube Music](https://music.youtube.com)._**\n\n## Installation\n\nRefer to our [Installation Guide](https://spotdl.rtfd.io/en/latest/installation/) for more details\n\n\n- Python (**Recommended**)\n  - _spotDL_ can be installed by running `pip install spotdl`.\n  > On some systems you might have to change `pip` to `pip3`.\n\n### Other options\n\n- Prebuilt Executable\n  - You can download the latest version from from the [Releases Tab](https://github.com/spotDL/spotify-downloader/releases)\n- On Termux\n  - `curl -L https://raw.githubusercontent.com/spotDL/spotify-downloader/master/termux/setup_spotdl.sh | sh`\n- Arch\n  - There is an Arch User Repository (AUR) package for [spotDL](https://aur.archlinux.org/packages/python-spotdl/).\n- Docker\n  - Build image:\n\n    ```bash\n    docker build -t spotdl .\n    ```\n\n  - Launch container with spotDL parameters (see section below). You need to create mapped volume to access song files\n\n    ```bash\n    docker run --rm -v $(pwd):/music spotdl download [trackUrl]\n    ```\n\n\n### Installing FFmpeg\n\nIf using FFmpeg only for spotDL, you can install FFmpeg to your local directory.  \n`spotdl --download-ffmpeg` will download FFmpeg to your spotDL installation directory.\n\nWe reccomend the above option, but if you want to install FFmpeg system-wide,\n\n- [Windows Tutorial](https://windowsloop.com/install-ffmpeg-windows-10/)\n- OSX - `brew install ffmpeg`\n- Linux - `sudo apt install ffmpeg` or use your distro\'s package manager\n\n## Usage\n\nTo get started right away:\n\n```sh\nspotdl download [urls]\n```\n\nTo start the Web UI:\n\n```sh\nspotdl web\n```\n\nYou can run _spotDL_ as a package if running it as a script doesn\'t work:\n\n```sh\npython -m spotdl [urls]\n```\n\n---\n\n### Further information can be found in our documentation\n\n**[Read the documentation on ReadTheDocs!](http://spotdl.rtfd.io/)**\n\n---\n\n## Contributing\n\nInterested in contributing? Check out our [CONTRIBUTING.md](docs/CONTRIBUTING.md) to find\nresources around contributing along with a guide on how to set up a development environment.\n\n## License\n\nThis project is Licensed under the [MIT](/LICENSE) License.\n',
    'author': 'spotDL Team',
    'author_email': 'spotdladmins@googlegroups.com',
    'maintainer': 'xnetcat',
    'maintainer_email': 'xnetcat.dev@gmail.com',
    'url': 'https://github.com/spotDL/spotify-downloader/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



"""
Module for embedding metadata into audio files using Mutagen.

```python
embed_metadata(
    output_file=Path("test.mp3"),
    song=song_object,
    file_format="mp3",
)
```
"""

import base64

from pathlib import Path
from urllib.request import urlopen

from mutagen.oggopus import OggOpus
from mutagen.mp4 import MP4, MP4Cover
from mutagen.flac import Picture, FLAC
from mutagen.oggvorbis import OggVorbis
from mutagen.easyid3 import EasyID3, ID3
from mutagen.id3 import APIC as AlbumCover, USLT, COMM as Comment

from spotdl.types import Song


class MetadataError(Exception):
    """
    Base class for all exceptions related to metadata and id3 embedding.
    """


# Apple has specific tags - see mutagen docs -
# http://mutagen.readthedocs.io/en/latest/api/mp4.html
M4A_TAG_PRESET = {
    "album": "\xa9alb",
    "artist": "\xa9ART",
    "date": "\xa9day",
    "title": "\xa9nam",
    "year": "\xa9day",
    "originaldate": "purd",
    "comment": "\xa9cmt",
    "group": "\xa9grp",
    "writer": "\xa9wrt",
    "genre": "\xa9gen",
    "tracknumber": "trkn",
    "albumartist": "aART",
    "discnumber": "disk",
    "cpil": "cpil",
    "albumart": "covr",
    "encodedby": "\xa9too",
    "copyright": "cprt",
    "tempo": "tmpo",
    "lyrics": "\xa9lyr",
    "explicit": "rtng",
}

TAG_PRESET = {key: key for key in M4A_TAG_PRESET}


def set_id3_mp3(output_file: Path, song: Song):
    """
    Set ID3 tags for MP3 files.

    ### Arguments
    - output_file: Path to the output file.
    - song: Song object.
    """

    audio_file = EasyID3(str(output_file.resolve()))
    audio_file.delete()

    audio_file["title"] = song.name
    audio_file["titlesort"] = song.name
    audio_file["tracknumber"] = [song.track_number, song.tracks_count]
    audio_file["discnumber"] = [song.disc_number, song.disc_count]
    audio_file["artist"] = song.artists
    audio_file["album"] = song.album_name
    audio_file["albumartist"] = song.album_artist
    audio_file["date"] = song.date
    audio_file["originaldate"] = song.date
    audio_file["encodedby"] = song.publisher
    if song.copyright_text:
        audio_file["copyright"] = song.copyright_text

    genres = song.genres
    if len(genres) > 0:
        audio_file["genre"] = genres[0]

    audio_file.save(v2_version=3)

    audio_file = ID3(str(output_file.resolve()))
    if song.cover_url:
        with urlopen(song.cover_url) as raw_album_art:
            audio_file["APIC"] = AlbumCover(
                encoding=3,
                mime="image/jpeg",
                type=3,
                desc="Cover",
                data=raw_album_art.read(),
            )

    if song.lyrics:
        audio_file["USLT::'eng'"] = USLT(
            encoding=3, lang="eng", desc="desc", text=song.lyrics
        )

    if song.download_url:
        audio_file.add(Comment(encoding=3, text=song.download_url))

    audio_file.save(v2_version=3)


def set_id3_m4a(output_file: Path, song: Song):
    """
    Set ID3 tags for M4A files.

    ### Arguments
    - output_file: Path to the output file.
    - song: Song object.
    """

    audio_file = MP4(str(output_file.resolve()))

    audio_file = _embed_basic_metadata(audio_file, song, "m4a", M4A_TAG_PRESET)

    audio_file[M4A_TAG_PRESET["year"]] = str(song.year)
    audio_file[M4A_TAG_PRESET["explicit"]] = (4 if song.explicit is True else 2,)

    if song.lyrics:
        audio_file[M4A_TAG_PRESET["lyrics"]] = song.lyrics

    if song.cover_url:
        try:
            with urlopen(song.cover_url) as raw_album_art:
                audio_file[M4A_TAG_PRESET["albumart"]] = [
                    MP4Cover(
                        raw_album_art.read(),
                        imageformat=MP4Cover.FORMAT_JPEG,
                    )
                ]
        except IndexError:
            pass

    if song.download_url:
        audio_file[M4A_TAG_PRESET["comment"]] = song.download_url

    audio_file.save()


def set_id3_flac(output_file: Path, song: Song):
    """
    Set ID3 tags for FLAC files.

    ### Arguments
    - output_file: Path to the output file.
    - song: Song object.
    """

    audio_file = FLAC(str(output_file.resolve()))

    audio_file = _embed_basic_metadata(audio_file, song, "flac")
    audio_file = _embed_ogg_metadata(audio_file, song)
    audio_file = _embed_cover(audio_file, song, "flac")

    audio_file.save()


def set_id3_opus(output_file: Path, song: Song):
    """
    Set ID3 tags for Opus files.

    ### Arguments
    - output_file: Path to the output file.
    - song: Song object.
    """

    audio_file = OggOpus(str(output_file.resolve()))

    audio_file = _embed_basic_metadata(audio_file, song, "opus")
    audio_file = _embed_ogg_metadata(audio_file, song)
    audio_file = _embed_cover(audio_file, song, "opus")

    audio_file.save()


def set_id3_ogg(output_file: Path, song: Song):
    """
    Set ID3 tags for OGG files.

    ### Arguments
    - output_file: Path to the output file.
    - song: Song object.
    """

    audio_file = OggVorbis(str(output_file.resolve()))

    audio_file = _embed_basic_metadata(audio_file, song, "ogg")
    audio_file = _embed_ogg_metadata(audio_file, song)
    audio_file = _embed_cover(audio_file, song, "ogg")

    audio_file.save()


def _embed_basic_metadata(audio_file, song: Song, encoding, preset=TAG_PRESET):
    """
    Embed basic metadata into the audio file.

    ### Arguments
    - audio_file: Mutagen audio file object.
    - song: Song object.
    - encoding: Encoding of the audio file.
    - preset: Preset of the audio file.

    ### Returns
    - Modified audio_file object.
    """

    album_name = song.album_name
    if album_name:
        audio_file[preset["album"]] = album_name

    audio_file[preset["artist"]] = song.artist
    audio_file[preset["albumartist"]] = song.artist
    audio_file[preset["title"]] = song.name
    audio_file[preset["date"]] = song.date
    audio_file[preset["originaldate"]] = song.date

    if len(song.genres) > 0:
        audio_file[preset["genre"]] = song.genres[0]

    if song.copyright_text:
        audio_file[preset["copyright"]] = song.copyright_text

    if encoding in ["flac", "ogg", "opus"]:
        zfilled_disc_number = str(song.disc_number).zfill(len(str(song.disc_count)))
        audio_file[preset["discnumber"]] = zfilled_disc_number
    else:
        audio_file[preset["discnumber"]] = [(song.disc_number, song.disc_count)]

    if encoding in ["flac", "ogg", "opus"]:
        zfilled_track_number = str(song.track_number).zfill(len(str(song.tracks_count)))
        audio_file[preset["tracknumber"]] = zfilled_track_number
    else:
        audio_file[preset["tracknumber"]] = [(song.track_number, song.tracks_count)]

    if encoding == "m4a":
        audio_file[preset["encodedby"]] = song.publisher

    return audio_file


def _embed_ogg_metadata(audio_file, song: Song):
    """
    Embed basic metadata into the ogg file.

    ### Arguments
    - audio_file: Mutagen audio file object.
    - song: Song object.

    ### Returns
    - Modified audio_file object.
    """

    audio_file["year"] = str(song.year)

    if song.lyrics:
        audio_file["lyrics"] = song.lyrics

    if song.download_url:
        audio_file["comment"] = song.download_url

    return audio_file


def _embed_cover(audio_file, song: Song, encoding: str):
    """
    Embed cover into the audio file.

    ### Arguments
    - audio_file: Mutagen audio file object.
    - song: Song object.
    - encoding: Encoding of the audio file.

    ### Returns
    - Modified audio_file object.
    """

    if song.cover_url is None:
        return audio_file

    image = Picture()
    image.type = 3
    image.desc = "Cover"
    image.mime = "image/jpeg"

    with urlopen(song.cover_url) as raw_album_art:
        image.data = raw_album_art.read()

    if encoding == "flac":
        audio_file.add_picture(image)
    elif encoding in ["ogg", "opus"]:
        image_data = image.write()
        encoded_data = base64.b64encode(image_data)
        vcomment_value = encoded_data.decode("ascii")
        audio_file["metadata_block_picture"] = [vcomment_value]

    return audio_file


AVAILABLE_FORMATS = {
    "mp3": set_id3_mp3,
    "flac": set_id3_flac,
    "opus": set_id3_opus,
    "ogg": set_id3_ogg,
    "m4a": set_id3_m4a,
}


def embed_metadata(output_file: Path, song: Song, file_format: str) -> None:
    """
    Embeds metadata into the output file.

    ### Arguments
    - output_file: Path to the output file.
    - song: Song object.
    - file_format: File format of the output file.
    """

    function = AVAILABLE_FORMATS.get(file_format)
    if function:
        function(output_file, song)
